"""This module defines Python equivalents for the enumerations found in the C library header file *dwf.h*.

Note:
    The obsolete enumerations *TRIGCOND* and *STS* that are defined in the C header file have not been defined here.
    *TRIGCOND* has been replaced by |DwfTriggerSlope:link|; *STS* has been replaced by |DwfState:link|.
"""

# pylint: disable= too-many-lines

import enum


class _FixReprMethod:
    """Mixin class that overrides the __repr__ methods for Enum classes.

    Enums in Python have a strange default '__repr__' implementation. This mixin class fixes that.
    """

    # pylint: disable=no-member, too-few-public-methods

    def __repr__(self):
        return "{}.{}".format(self.__class__.__name__, self.name)


@enum.unique
class DwfErrorCode(_FixReprMethod, enum.Enum):
    """Enumeration type for error reporting constants of the DWF API.

    This type is used by the |DwfLibrary.getLastError:link| method to report the error condition of the most
    recent C API call.

    In |pydwf|, it is only used as the type of the :py:attr:`~pydwf.core.auxiliary.exceptions.DwfLibraryError.code`
    field of |DwfLibraryError:link| instances.

    In the C API, this type is called 'DWFERC', and it is represented as an *int*.
    """

    NoErc = 0
    """No error occurred."""

    UnknownError = 1
    """Unknown error."""

    ApiLockTimeout = 2
    """API waiting on pending operation timed out."""

    AlreadyOpened = 3
    """Device already opened."""

    NotSupported = 4
    """Device not supported."""

    InvalidParameter0 = 0x10
    """Invalid parameter sent in API call."""

    InvalidParameter1 = 0x11
    """Invalid parameter sent in API call."""

    InvalidParameter2 = 0x12
    """Invalid parameter sent in API call."""

    InvalidParameter3 = 0x13
    """Invalid parameter sent in API call."""

    InvalidParameter4 = 0x14
    """Invalid parameter sent in API call."""


class DwfEnumFilter(_FixReprMethod, enum.Flag):
    """Enumeration type for device class constants, used during device enumeration.

    This type is used by the |DeviceEnum.enumerateDevices:link| and |DeviceEnum.enumerateStart:link|
    methods to constrain the type of devices that will be enumerated.

    In the C API, this type is called 'ENUMFILTER', and it is represented as an *int*.
    """

    All = 0
    """Enumerate all available devices."""

    Type = 0x08000000
    """Enumerate by type (0x08000000)."""

    USB = 0x00000001
    """Enumerate USB devices."""

    Network = 0x00000002
    """Enumerate Network devices."""

    AXI = 0x00000004
    """Enumerate AXI devices (used when running on an ADP 3x50 device)."""

    Remote = 0x01000000
    """Enumerate remote devices (0x01000000)."""

    Audio = 0x02000000
    """Enumerate audio devices (0x02000000)."""

    Demo = 0x04000000
    """Enumerate demo devices (0x04000000)."""


@enum.unique
class DwfEnumConfigInfo(_FixReprMethod, enum.Enum):
    """Enumeration type for device configuration parameter type constants.

    This type lists the device parameters that can vary between different |device configurations:link|
    of the same device. It is used exclusively by the |DeviceEnum.configInfo:link| method.

    In the C API, this type is represented as an *int*.
    """

    #pylint: disable=line-too-long

    TooltipText = -1
    """Tooltip text.

    Maximum length: 2048 characters.

    Note:
        This value is not officially documented. Its existence was revealed in a
        `message on the Digilent forum <https://forum.digilentinc.com/topic/21720-small-issue-and-questions-about-device-configurations/#comment-62717>`_.
    """

    OtherInfoText = -2
    """Other info text.

    Maximum length: 256 characters.

    Note:
        This value is not officially documented. Its existence was revealed in a
        `message on the Digilent forum <https://forum.digilentinc.com/topic/21720-small-issue-and-questions-about-device-configurations/#comment-62717>`_.
    """

    AnalogInChannelCount = 1
    """Number of analog input channels."""

    AnalogOutChannelCount = 2
    """Number of analog output channels."""

    AnalogIOChannelCount = 3
    """Number of analog power supply channels.

    Note:
        This is a different number than the number of channels reported by the |AnalogIO.channelCount:link|
        method.
    """

    DigitalInChannelCount = 4
    """Number of digital input channels."""

    DigitalOutChannelCount = 5
    """Number of digital output channels."""

    DigitalIOChannelCount = 6
    """Number of digital I/O channels."""

    AnalogInBufferSize = 7
    """Analog in buffer size, in samples."""

    AnalogOutBufferSize = 8
    """Analog out buffer size, in samples."""

    DigitalInBufferSize = 9
    """Digital in buffer size, in samples."""

    DigitalOutBufferSize = 10
    """Digital out buffer size, in samples."""


@enum.unique
class DwfDeviceID(_FixReprMethod, enum.Enum):
    """Enumeration type for device ID constants.

    This type is used by the |DeviceEnum.deviceType:link| method to report the type of a previously enumerated device.

    In the C API, this type is called 'DEVID', and it is represented as an *int*.
    """

    EExplorer = 1
    """Electronics Explorer devices."""

    Discovery = 2
    """Analog Discovery devices."""

    Discovery2 = 3
    """Analog Discovery 2 devices."""

    DDiscovery = 4
    """Digital Discovery devices."""

    ADP3X50 = 6
    """Analog Discovery Pro devices."""

    Eclypse = 7
    """Eclypse devices."""

    ADP5250 = 8
    """ADP5250 devices."""

    DPS3340 = 9
    """DPS 3340 devices."""


class DwfDeviceVersion(_FixReprMethod, enum.Enum):
    """Enumeration type for device version (i.e., hardware revision) constants.

    This type is used by the |DeviceEnum.deviceType:link| method to report the
    hardware revision of a previously enumerated device.

    Note:
        The device revision list given here is not complete; it does not cover all devices.

    In the C API, this type is called 'DEVVER', and it is represented as an *int*.
    """

    EExplorerC = 2
    """Electronics Explorer devices, revision C."""
    EExplorerE = 4
    """Electronics Explorer devices, revision E."""
    EExplorerF = 5
    """Electronics Explorer devices, revision F."""
    DiscoveryA = 1
    """Discovery devices, revision A."""
    DiscoveryB = 2
    """Discovery devices, revision B."""
    DiscoveryC = 3
    """Discovery devices, revision C."""


@enum.unique
class DwfDeviceParameter(_FixReprMethod, enum.Enum):
    """Enumeration type for device parameter constants.

    Device parameters are miscellaneous integer settings that influence the behavior of a device.

    The different device parameters are selected by one of the constant values defined here.

    This type is used to select device parameters, either to set/get global defaults using the |DwfLibrary|, or to
    to set/get parameter values on a specific, previously opened device |DwfDevice|.

    In the C API, this type is called 'DwfParam', and it is represented as an *int*.
    """

    KeepOnClose = 1
    """Keep the device running after close.

    Warning:

        This value is *obsolete*. Use *OnClose* instead.
    """

    UsbPower = 2
    """USB power behavior if AUX power is connected.

    Possible values:

    * 0 — Disable USB power.
    * 1 — Keep USB power enabled.

    This setting is implemented on the Analog Discovery 2.
    """

    LedBrightness = 3
    """Set multicolor LED brightness.

    The Digital Discovery features a multi-color LED. It is normally blue in case the device is not
    currently controlled by software, or green if it is.

    Setting this parameter from 0 to 100 changes the LED's relative brightness, in percents. This can be
    useful, for example, in a lab with sensitive optics that would preferably be completely dark.

    On the Analog Discovery 2, this setting has no effect.
    """

    OnClose = 4
    """Define behavior on close.

    Possible values:

    * 0 — On close, continue.
    * 1 — On close, stop the device.
    * 2 — On close, shut down the device.
    """

    AudioOut = 5
    """Enable or disable audio output.

    Possible values:

    * 0 — Disable audio output.
    * 1 — Enable audio output.

    This setting is implemented on the Analog Discovery and the Analog Discovery 2.
    """

    UsbLimit = 6
    """USB power limit.

    The value ranges from 0 to 1000, in mA. The value -1 denotes no limit.

    This setting is implemented on the Analog Discovery and the Analog Discovery 2.
    """

    AnalogOut = 7
    """Enable or disable analog output.

    Possible values:

    * 0 — Disable analog output.
    * 1 — Enable analog output.
    """

    Frequency = 8
    """This parameter is undocumented.

    Todo:
        The meaning of this parameter needs to be understood.

        It is some frequency; comments say unit is Hz.
    """

    ExtFreq = 9
    """This parameter is undocumented.

    Todo:
        The meaning of this parameter needs to be understood.

        It is some frequency; comments say unit is Hz.
    """

    ClockMode = 10
    """This parameter is undocumented.

    Todo:
        The meaning of this parameter needs to be understood.

    Possible values:

    * 0 — internal.
    * 1 — output.
    * 2 — input.
    * 3 — IO.
    """

    TempLimit = 11
    """This parameter is undocumented.

    Todo:
        The meaning of this parameter needs to be understood.
    """

    FreqPhase = 12
    """This parameter is undocumented.

    Todo:
        The meaning of this parameter needs to be understood.
    """


@enum.unique
class DwfWindow(_FixReprMethod, enum.Enum):
    """Enumeration type for signal processing windows.
    """

    Rectangular = 0
    """Rectangular window, a.k.a. no window."""

    Triangular = 1
    """Triangular window."""

    Hamming = 2
    """Hamming window."""

    Hann = 3
    """Hann window."""

    Cosine = 4
    """Cosine window."""

    BlackmanHarris = 5
    """Blackman-Harris window."""

    FlatTop = 6
    """Flat-top window."""

    Kaiser = 7
    """Kaiser window."""


class DwfState(_FixReprMethod, enum.Enum):
    """Enumeration type for instrument state constants, for instruments that are controlled by an internal
    state-machine.

    The following instrument APIs are controlled by a state machine:

    * |AnalogIn|
    * |AnalogOut| — *independent state machine for each channel*
    * |DigitalIn|
    * |DigitalOut|
    * |AnalogImpedance|

    This type is used to return the current state from their *status()* methods: |AnalogIn.status:link|,
    |AnalogOut.status:link|, |DigitalIn.status:link|, |DigitalOut.status:link|, and |AnalogImpedance.status:link|.

    Note:
        The enumeration values :py:attr:`Triggered` and :py:attr:`Running` have identical integer values (3).

        The state name :py:attr:`Triggered` is used for capture instruments (|AnalogIn|, |DigitalIn|),
        while :py:attr:`Running` is used for signal generation instruments (|AnalogOut|, |DigitalOut|).

    In the C API, this type is represented as an *unsigned char*.
    """

    Ready = 0
    """The instrument is idle, waiting to be configured or started."""

    Config = 4
    """The instrument is being configured."""

    Prefill = 5
    """The instrument is collecting data prior to arming itself, so it can deliver pre-trigger samples."""

    Armed = 1
    """The instrument is collecting samples and waiting for the trigger."""

    Wait = 7
    """The signal generation instrument is waiting before its next run."""

    Triggered = 3
    """The capture instrument is triggered and collecting data."""

    Running = 3
    """The signal generation instrument is running (generating signals)."""

    Done = 2
    """The instrument has completed a measurement or signal-generating sequence."""


@enum.unique
class DwfTriggerSource(_FixReprMethod, enum.Enum):
    """Enumeration type for trigger source constants.

    This type is used by the |DeviceControl| functionality to configure an external trigger and by the
    |AnalogIn|, |AnalogOut|, |DigitalIn|, and |DigitalOut| instruments to select a trigger source.
    The |AnalogIn| instrument can also use this type to configure a trigger as a sampling source.

    In the C API, this type is called 'TRIGSRC', and it is represented as an *unsigned char*.
    """

    None_ = 0
    """No trigger configured (device starts immediately)."""

    PC = 1
    """PC (software) trigger."""

    DetectorAnalogIn  = 2
    """AnalogIn trigger detector."""

    DetectorDigitalIn = 3
    """DigitalIn trigger detector."""

    AnalogIn = 4
    """AnalogIn instrument start."""

    DigitalIn = 5
    """DigitalIn instrument start."""

    DigitalOut = 6
    """DigitalOut instrument start."""

    AnalogOut1 = 7
    """|AnalogOut| instrument channel 1 start."""

    AnalogOut2 = 8
    """|AnalogOut| instrument channel 2 start."""

    AnalogOut3 = 9
    """|AnalogOut| instrument channel 3 start."""

    AnalogOut4 = 10
    """|AnalogOut| instrument channel 4 start."""

    External1 = 11
    """External trigger #1."""

    External2 = 12
    """External trigger #2."""

    External3 = 13
    """External trigger #3."""

    External4 = 14
    """External trigger #4."""

    High = 15
    """High (undocumented)."""

    Low = 16
    """Low (undocumented)."""

    Clock = 17
    """Clock (undocumented)."""


@enum.unique
class DwfTriggerSlope(_FixReprMethod, enum.Enum):
    """Enumeration type for trigger slope constants.

    This type is used by the |AnalogIn|, |AnalogOut|, |DigitalIn|, and |DigitalOut| instruments to select
    the trigger slope.

    In addition, the |AnalogIn| instrument uses it to select the slope of the sampling clock.

    In the C API, this type is represented as an *int*.
    """

    Rise = 0
    """Rising trigger slope."""

    Fall = 1
    """Falling trigger slope."""

    Either = 2
    """Either rising or falling trigger slope."""


@enum.unique
class DwfAcquisitionMode(_FixReprMethod, enum.Enum):
    """Enumeration type for acquisition mode constants.

    This type is used by the |AnalogIn| and |DigitalIn| instruments. These instruments
    support multiple acquisition modes that are appropriate for different data
    acquisition tasks.

    In the C API, this type is called 'ACQMODE', and it is represented as an *int*.
    """

    Single = 0
    """Perform a single buffer acquisition.

    Re-arm the instrument for the next capture after the data is fetched to the host
    using the instrument-specific *status()* method: |AnalogIn.status:link| or
    |DigitalIn.status:link|.

    Note:
        The difference with the :py:attr:`Single1` mode is unclear.
    """

    ScanShift = 1
    """Perform a continuous acquisition in FIFO style.

    The trigger setting is ignored.

    The last sample is at the end of the buffer.
    The instrument's *statusSamplesValid()* method gives the number of the acquired samples,
    which will increase until reaching the buffer size.
    After that, the waveform image is shifted for every new sample.
    """

    ScanScreen = 2
    """Perform continuous acquisition circularly writing samples into the buffer.

    This is similar to a heart-monitor display.

    The trigger setting is ignored.

    The instrument's *statusIndexWrite()* method gives the buffer write position.
    """

    Record = 3
    """Perform acquisition for the length of time set by the instrument's *recordLengthSet()* method."""

    Overs = 4
    """Overscan mode (undocumented)."""

    Single1 = 5
    """Perform a single buffer acquisition.

    Note:
        The difference with the :py:attr:`Single` mode is unclear.
    """


@enum.unique
class DwfAnalogInFilter(_FixReprMethod, enum.Enum):
    """Enumeration type for analog input filter constants.

    This type is used by the |AnalogIn| instrument to select a filtering algorithm for the input and trigger channels.

    The |AnalogIn| instrument's ADC always captures samples at the maximum possible rate. If data acquisition at
    a lower sampling rate is requested, the resampling can be handled in several different ways.

    The most obvious choice is *averaging*. This will suppress high-frequency noise, which is often a good thing,
    but sometimes it is desirable to know that high-frequency noise is present in the signal, and the averaging
    may hide that fact.

    For that reason, the *decimation* filter is available, which simply selects a single sample captured at high
    frequency when resampling to a lower frequency. The signal-to-noise ratio (SNR) will suffer, but the presence
    of high-frequency noise (outliers) will be more easily seen in the resampled data.

    Todo:
        Examine the MinMax filter choice; it is not currently understood.

    In the C API, this type is called 'FILTER', and it is represented as an *int*.
    """

    Decimate = 0
    """Decimation filter."""

    Average = 1
    """Averaging filter."""

    MinMax = 2
    """Min/max filter."""

    AverageFit = 3
    """Averaging fit filter."""


@enum.unique
class DwfAnalogCoupling(_FixReprMethod, enum.Enum):
    """Enumeration type for analog coupling configuration.
    """

    DC = 0
    """DC coupling."""

    AC = 1
    """AC coupling."""


@enum.unique
class DwfAnalogInTriggerType(_FixReprMethod, enum.Enum):
    """Enumeration type for analog input trigger mode constants.

    This type is used by the |AnalogIn| instrument to specify the trigger type.

    In the C API, this type is called 'TRIGTYPE', and it is represented as an *int*.
    """

    Edge = 0
    """Edge trigger type."""

    Pulse = 1
    """Pulse trigger type."""

    Transition = 2
    """Transition trigger type."""

    Window = 3
    """Window trigger type."""


@enum.unique
class DwfAnalogInTriggerLengthCondition(_FixReprMethod, enum.Enum):
    """Enumeration type for analog input trigger length condition constants.

    This type is used by the |AnalogIn| instrument to specify the trigger length condition.

    In the C API, this type is called 'TRIGLEN', and it is represented as an *int*.
    """

    Less = 0
    """Trigger length condition 'less'."""

    Timeout = 1
    """Trigger length condition 'timeout'."""

    More = 2
    """Trigger length condition 'more'."""


@enum.unique
class DwfAnalogOutFunction(_FixReprMethod, enum.Enum):
    """Enumeration type for analog output waveform-shape function constants.

    This type is used by the |AnalogOut| instrument to represent the wave-shape produced on an
    analog output channel node. The nine fixed waveform shape options are shown below.

    .. only:: html

        .. image:: /images/waveforms.gif

    .. only:: latex

        .. image:: /images/waveforms.pdf

    In the C API, this type is called 'FUNC', and it is represented as an *unsigned char*.
    """

    DC = 0
    """DC (constant signal) waveform shape. The signal level varies between -1 and 1."""

    Sine = 1
    """Sinusoid waveform shape. The signal level varies between -1 and 1."""

    Square = 2
    """Square waveform shape. The signal level varies between -1 and 1."""

    Triangle = 3
    """Triangle waveform shape. The signal level varies between -1 and 1."""

    RampUp = 4
    """Ramp Up waveform shape. The signal level varies between -1 and 1."""

    RampDown = 5
    """Ramp Down waveform shape. The signal level varies between -1 and 1."""

    Noise = 6
    """Noise waveform shape. The signal level is uniformly distributed between -1 and 1."""

    Pulse = 7
    """Pulse waveform shape. The signal level varies between 0 and 1."""

    Trapezium = 8
    """Trapezium waveform shape. The signal level varies between -1 and 1."""

    SinePower = 9
    """Sinusoid Power waveform shape. The signal level varies between -1 and 1."""

    CustomPattern = 28
    """(Not documented.)"""

    PlayPattern = 29
    """(Not documented.)"""

    Custom = 30
    """Custom (user-defined) waveform shape. The signal level varies between -1 and 1."""

    Play = 31
    """Continuous playback mode. The signal level varies between -1 and 1."""

    AnalogIn1 = 64
    """(Not documented.)"""

    AnalogIn2 = 65
    """(Not documented.)"""

    AnalogIn3 = 66
    """(Not documented.)"""

    AnalogIn4 = 67
    """(Not documented.)"""

    AnalogIn5 = 68
    """(Not documented.)"""

    AnalogIn6 = 69
    """(Not documented.)"""

    AnalogIn7 = 70
    """(Not documented.)"""

    AnalogIn8 = 71
    """(Not documented.)"""

    AnalogIn9 = 72
    """(Not documented.)"""

    AnalogIn10 = 73
    """(Not documented.)"""

    AnalogIn11 = 74
    """(Not documented.)"""

    AnalogIn12 = 75
    """(Not documented.)"""

    AnalogIn13 = 76
    """(Not documented.)"""

    AnalogIn14 = 77
    """(Not documented.)"""

    AnalogIn15 = 78
    """(Not documented.)"""

    AnalogIn16 = 79
    """(Not documented.)"""


@enum.unique
class DwfAnalogOutNode(_FixReprMethod, enum.Enum):
    """Enumeration type for analog output node type constants.

    This type is used by the |AnalogOut| instrument to represent the node types associated with each output channel.

    In the C API, this type is called 'AnalogOutNode' (without the *Dwf* prefix), and it is represented as an *int*.
    """

    Carrier = 0
    """Carrier signal node. This node represents the base signal without modulation applied."""

    FM = 1
    """Frequency Modulation node."""

    AM = 2
    """Amplitude Modulation node."""


@enum.unique
class DwfAnalogOutMode(_FixReprMethod, enum.Enum):
    """Enumeration type for analog out mode constants (voltage or current).

    This type is used by the |AnalogOut| instrument to set or retrieve the mode of a channel.

    In the C API, this type is represented as an *int*.
    """

    Voltage = 0
    """Voltage mode."""

    Current = 1
    """Current mode."""


@enum.unique
class DwfAnalogOutIdle(_FixReprMethod, enum.Enum):
    """Enumeration type for analog output idle state constants.

    This type is used by the |AnalogOut| instrument to set the idle behavior of an output channel.

    In the C API, this type is represented as an *int*.
    """

    Disable = 0
    """When idle, disable the output."""

    Offset = 1
    """When idle, drive the analog output offset."""

    Initial = 2
    """When idle, drive the initial value of the selected waveform shape."""


@enum.unique
class DwfDigitalInClockSource(_FixReprMethod, enum.Enum):
    """Enumeration type for digital input clock source constants.

    This type is used by the |DigitalIn| instrument to specify a clock source.

    In the C API, this type is represented as an *int*.
    """

    Internal = 0
    """Use internal clock source."""

    External = 1
    """Use external clock source."""

    External2 = 2
    """Use alternate external clock source."""


@enum.unique
class DwfDigitalInSampleMode(_FixReprMethod, enum.Enum):
    """Enumeration type for digital input sample mode constants.

    This type is used by the |DigitalIn| instrument to specify a sample mode.

    In the C API, this type is represented as an *int*.
    """

    Simple = 0
    """Only digital samples (no noise)."""

    Noise = 1
    """Alternate samples (noise, sample, noise, sample, …) where noise is more than one transition between two samples.

    This setting is available when the sample rate is less than the maximum clock frequency (i.e., the divider is
    greater than one). Digital noise means more than one transition between subsequent samples was detected; this
    can indicate glitches or ringing.
    """


@enum.unique
class DwfDigitalOutOutput(_FixReprMethod, enum.Enum):
    """Enumeration type for digital output mode constants.

    This type is used by the |DigitalOut| instrument to specify the electronic behavior of a digital output channel.

    In the C API, this type is represented as an *int*.
    """

    PushPull = 0
    """Push/Pull."""

    OpenDrain = 1
    """Open Drain."""

    OpenSource = 2
    """Open Source."""

    ThreeState = 3
    """Tristate (for custom and random)."""


@enum.unique
class DwfDigitalOutType(_FixReprMethod, enum.Enum):
    """Enumeration type for digital output type constants.

    This type is used by the |DigitalOut| instrument to specify the behavior mode of a digital output channel.

    In the C API, this type is represented as an *int*.
    """

    Pulse = 0
    """Pulse output."""

    Custom = 1
    """Custom output."""

    Random = 2
    """Random output."""

    ROM = 3
    """ROM (lookup table) output."""

    State = 4
    """State machine output."""

    Play = 5
    """Continuous playback output."""


@enum.unique
class DwfDigitalOutIdle(_FixReprMethod, enum.Enum):
    """Enumeration type for digital output idle mode constants.

    This type is used primarily by the |DigitalOut| instrument to specify the idle behavior mode of a digital
    output channel.

    In addition to that, it is used by the |ProtocolSPI| functionality to specify the idle behavior
    of the pins it controls.

    In the C API, this type is represented as an *int*.
    """

    Init = 0
    """Same as initial value of selected output pattern."""

    Low  = 1
    """Low signal level."""

    High = 2
    """High signal level."""

    Zet  = 3
    """High impedance."""


@enum.unique
class DwfAnalogIO(_FixReprMethod, enum.Enum):
    """Enumeration type for Analog I/O channel node type constants.

    This type is used by the |AnalogIO| functionality to report the node type.

    In the C API, this type is called 'ANALOGIO', and it is represented as an *unsigned char*.
    """

    Undocumented = 0
    """This value was returned in Analog Discovery Pro devices when using 3.16.3 of the DWF library.
       This was a bug that was subsequently fixed."""

    Enable = 1
    """The node represent an on/off switch."""

    Voltage = 2
    """The node represents a voltage."""

    Current = 3
    """The node represents a current."""

    Power = 4
    """The node represents a power."""

    Temperature = 5
    """The node represents a temperature."""

    Dmm = 6
    """The node represents a DMM (digital multimeter) value."""

    Range = 7
    """The node represents a range."""

    Measure = 8
    """(unknown)"""

    Time = 9
    """The node represents a time."""

    Frequency = 10
    """The node represents a frequency."""

    Resistance = 11
    """The node represents a resistance."""

    Slew = 12
    """(unknown)"""


@enum.unique
class DwfAnalogImpedance(_FixReprMethod, enum.Enum):
    """Enumeration type for analog impedance measurement types.

    This type is used by the |AnalogImpedance| measurement functionality to specify a measurement quantity type.

    In the C API, this type is represented as an *int*.
    """

    Impedance = 0
    """Measure impedance, in Ohms."""

    ImpedancePhase = 1
    """Measure impedance phase, in radians."""

    Resistance = 2
    """Measure resistance, in Ohms."""

    Reactance = 3
    """Measure reactance, in Ohms."""

    Admittance = 4
    """Measure admittance, in Siemens."""

    AdmittancePhase = 5
    """Measure admittance phase, in radians."""

    Conductance = 6
    """Measure conductance, in Siemens."""

    Susceptance = 7
    """Measure susceptance, in Siemens."""

    SeriesCapacitance = 8
    """Measure series capacitance, in Farad."""

    ParallelCapacitance = 9
    """Measure parallel capacitance, in Farad."""

    SeriesInductance = 10
    """Measure series inductance, in Henry."""

    ParallelInductance = 11
    """Measure parallel inductance, in Henry."""

    Dissipation = 12
    """Measure dissipation, as a factor."""

    Quality = 13
    """Measure quality, as a factor."""

    Vrms = 14
    """Measure Vrms, in Volts."""

    Vreal = 15
    """Measure Vreal (real part of complex voltage), in Volts."""

    Vimag = 16
    """Measure Vimag (imaginary part of complex voltage), in Volts."""

    Irms = 17
    """Measure Irms, in Amps."""

    Ireal = 18
    """Measure Ireal (real part of complex current), in Amps."""

    Iimag = 19
    """Measure Iimag (imaginary part of complex current), in Amps."""


@enum.unique
class DwfDmm(_FixReprMethod, enum.Enum):
    """Enumeration type for DMM (digital multimeter) measurements.

    Note:

        This type is currently unused in the API. It is intended for functionality in the new ADP5250 device.

    In the C API, this type is called 'DwfDmm', and it is represented as an *int*.
    """

    Resistance = 1
    """Resistance measurement."""

    Continuity = 2
    """Continuity measurement."""

    Diode = 3
    """Diode measurement."""

    DCVoltage = 4
    """DC voltage measurement."""

    ACVoltage = 5
    """AC voltage measurement."""

    DCCurrent = 6
    """DC current measurement."""

    ACCurrent = 7
    """AC current measurement."""

    DCLowCurrent = 8
    """DC low current measurement."""

    ACLowCurrent = 9
    """AC low current measurement."""

    Temperature = 10
    """Temperature measurement."""



"""This module provides the :py:class:`typespec_ctypes` class.

It is used to map the C types found in the 'dwf.h' header file to specific *ctypes* types.
"""

import ctypes


class typespec_ctypes:
    """Map the type specifications from :py:mod:`pydwf.core.auxiliary.dwf_function_signatures` to *ctypes* types."""

    # pylint: disable=too-few-public-methods

    # The basic C types.
    #
    # Note: the use of type 'bool' in dwf.h is probably not intentional;
    #  it is not a C type (unless stdbool.h bas been included).

    c_bool                                = ctypes.c_bool

    c_char                                = ctypes.c_char
    c_char_ptr                            = ctypes.POINTER(c_char)
    c_const_char_ptr                      = ctypes.c_char_p

    c_char_array_16                       = c_char * 16
    c_char_array_32                       = c_char * 32
    c_char_array_512                      = c_char * 512

    c_short                               = ctypes.c_short
    c_short_ptr                           = ctypes.POINTER(c_short)

    c_int                                 = ctypes.c_int
    c_int_ptr                             = ctypes.POINTER(c_int)

    c_unsigned_char                       = ctypes.c_ubyte
    c_unsigned_char_ptr                   = ctypes.POINTER(c_unsigned_char)

    c_unsigned_short                      = ctypes.c_ushort
    c_unsigned_short_ptr                  = ctypes.POINTER(c_unsigned_short)

    c_unsigned_int                        = ctypes.c_uint
    c_unsigned_int_ptr                    = ctypes.POINTER(c_unsigned_int)

    c_unsigned_long_long                  = ctypes.c_ulonglong
    c_unsigned_long_long_ptr              = ctypes.POINTER(c_unsigned_long_long)

    c_double                              = ctypes.c_double
    c_double_ptr                          = ctypes.POINTER(c_double)
    c_double_array_32                     = c_double * 32

    c_const_double                        = c_double
    c_const_double_ptr                    = ctypes.POINTER(c_const_double)

    c_void_ptr                            = ctypes.c_void_p

    # the HDWF type represents an open device handle.

    HDWF                                  = c_int
    HDWF_ptr                              = ctypes.POINTER(HDWF)

    # The 24 enum types follow; some of them were renamed from their C counterparts.

    DwfErrorCode                          = c_int
    DwfErrorCode_ptr                      = ctypes.POINTER(DwfErrorCode)

    DwfEnumFilter                         = c_int

    DwfEnumConfigInfo                     = c_int

    DwfDeviceID                           = c_int
    DwfDeviceID_ptr                       = ctypes.POINTER(DwfDeviceID)

    DwfDeviceVersion                      = c_int
    DwfDeviceVersion_ptr                  = ctypes.POINTER(DwfDeviceVersion)

    DwfDeviceParameter                    = c_int

    DwfWindow                             = c_int
    DwfWindow_ptr                         = ctypes.POINTER(DwfWindow)

    DwfState                              = c_unsigned_char
    DwfState_ptr                          = ctypes.POINTER(DwfState)

    DwfTriggerSource                      = c_unsigned_char
    DwfTriggerSource_ptr                  = ctypes.POINTER(DwfTriggerSource)

    DwfTriggerSlope                       = c_int
    DwfTriggerSlope_ptr                   = ctypes.POINTER(DwfTriggerSlope)

    DwfAcquisitionMode                    = c_int
    DwfAcquisitionMode_ptr                = ctypes.POINTER(DwfAcquisitionMode)

    DwfAnalogInFilter                     = c_int
    DwfAnalogInFilter_ptr                 = ctypes.POINTER(DwfAnalogInFilter)

    DwfAnalogInTriggerType                = c_int
    DwfAnalogInTriggerType_ptr            = ctypes.POINTER(DwfAnalogInTriggerType)

    DwfAnalogInTriggerLengthCondition     = c_int
    DwfAnalogInTriggerLengthCondition_ptr = ctypes.POINTER(DwfAnalogInTriggerLengthCondition)

    DwfAnalogCoupling                     = c_int
    DwfAnalogCoupling_ptr                 = ctypes.POINTER(DwfAnalogCoupling)

    DwfAnalogOutFunction                  = c_unsigned_char
    DwfAnalogOutFunction_ptr              = ctypes.POINTER(DwfAnalogOutFunction)

    DwfAnalogOutNode                      = c_int

    DwfAnalogOutMode                      = c_int
    DwfAnalogOutMode_ptr                  = ctypes.POINTER(DwfAnalogOutMode)

    DwfAnalogOutIdle                      = c_int
    DwfAnalogOutIdle_ptr                  = ctypes.POINTER(DwfAnalogOutIdle)

    DwfDigitalInClockSource               = c_int
    DwfDigitalInClockSource_ptr           = ctypes.POINTER(DwfDigitalInClockSource)

    DwfDigitalInSampleMode                = c_int
    DwfDigitalInSampleMode_ptr            = ctypes.POINTER(DwfDigitalInSampleMode)

    DwfDigitalOutOutput                   = c_int
    DwfDigitalOutOutput_ptr               = ctypes.POINTER(DwfDigitalOutOutput)

    DwfDigitalOutType                     = c_int
    DwfDigitalOutType_ptr                 = ctypes.POINTER(DwfDigitalOutType)

    DwfDigitalOutIdle                     = c_int
    DwfDigitalOutIdle_ptr                 = ctypes.POINTER(DwfDigitalOutIdle)

    DwfAnalogIO                           = c_unsigned_char
    DwfAnalogIO_ptr                       = ctypes.POINTER(DwfAnalogIO)

    DwfAnalogImpedance                    = c_int



"""This module provides convenience functions for configuration of instruments, channels, and nodes.

Note:
    This is a preview of the configuration API. It is not complete, has not fully stabilized yet, and it may change
    in future releases of |pydwf|. Currently, only the |AnalogOut| and |AnalogIn| instruments are supported.

    Future releases of |pydwf| will add support for more instruments.
"""

from typing import Optional

import numpy as np

from pydwf.core.api.analog_in import AnalogIn
from pydwf.core.api.analog_out import AnalogOut

from pydwf.core.auxiliary.enum_types import (DwfAnalogInFilter, DwfTriggerSource, DwfTriggerSlope,
                                             DwfAnalogInTriggerType, DwfAnalogInTriggerLengthCondition,
                                             DwfAcquisitionMode, DwfAnalogOutMode, DwfAnalogOutIdle, DwfAnalogOutNode,
                                             DwfAnalogOutFunction)

#######################################################################################################################
##                                                                                                                   ##
##                        Configuration support for the AnalogIn instrument and its channels                         ##
##                                                                                                                   ##
#######################################################################################################################

def configure_analog_in_instrument(
        analog_in                         : AnalogIn,
        acquisition_mode                  : Optional[DwfAcquisitionMode],
        sample_frequency                  : Optional[float]=None,
        buffer_size                       : Optional[int]=None,
        noise_buffer_size                 : Optional[int]=None,
        record_duration                   : Optional[float]=None,
        trigger_source                    : Optional[DwfTriggerSource]=None,
        trigger_position                  : Optional[float]=None,
        trigger_auto_timeout              : Optional[float]=None,
        trigger_detector_holdoff          : Optional[float]=None,
        trigger_detector_type             : Optional[DwfAnalogInTriggerType]=None,
        trigger_detector_channel          : Optional[int]=None,
        trigger_detector_filter           : Optional[DwfAnalogInFilter]=None,
        trigger_detector_level            : Optional[float]=None,
        trigger_detector_hysteresis       : Optional[float]=None,
        trigger_detector_condition        : Optional[DwfTriggerSlope]=None,
        trigger_detector_length           : Optional[float]=None,
        trigger_detector_length_condition : Optional[DwfAnalogInTriggerLengthCondition]=None,
        sampling_source                   : Optional[DwfTriggerSource]=None,
        sampling_slope                    : Optional[DwfTriggerSlope]=None,
        sampling_delay                    : Optional[float]=None) -> None:

    """Configure analog-input instrument acquisition, triggering, and sampling settings."""

    # pylint: disable=too-many-locals, too-many-branches

    if acquisition_mode is not None:
        analog_in.acquisitionModeSet(acquisition_mode)

    if sample_frequency is not None:
        analog_in.frequencySet(sample_frequency)

    if buffer_size is not None:
        analog_in.bufferSizeSet(buffer_size)

    if noise_buffer_size is not None:
        analog_in.noiseSizeSet(noise_buffer_size)

    if record_duration is not None:
        analog_in.recordLengthSet(record_duration)

    if trigger_source is not None:
        analog_in.triggerSourceSet(trigger_source)

    if trigger_position is not None:
        analog_in.triggerPositionSet(trigger_position)

    if trigger_auto_timeout is not None:
        analog_in.triggerAutoTimeoutSet(trigger_auto_timeout)

    if trigger_detector_holdoff is not None:
        analog_in.triggerHoldOffSet(trigger_detector_holdoff)

    if trigger_detector_type is not None:
        analog_in.triggerTypeSet(trigger_detector_type)

    if trigger_detector_channel is not None:
        analog_in.triggerChannelSet(trigger_detector_channel)

    if trigger_detector_filter is not None:
        analog_in.triggerFilterSet(trigger_detector_filter)

    if trigger_detector_level is not None:
        analog_in.triggerLevelSet(trigger_detector_level)

    if trigger_detector_hysteresis is not None:
        analog_in.triggerHysteresisSet(trigger_detector_hysteresis)

    if trigger_detector_condition is not None:
        analog_in.triggerConditionSet(trigger_detector_condition)

    if trigger_detector_length is not None:
        analog_in.triggerLengthSet(trigger_detector_length)

    if trigger_detector_length_condition is not None:
        analog_in.triggerLengthConditionSet(trigger_detector_length_condition)

    if sampling_source is not None:
        analog_in.samplingSourceSet(sampling_source)

    if sampling_slope is not None:
        analog_in.samplingSlopeSet(sampling_slope)

    if sampling_delay is not None:
        analog_in.samplingDelaySet(sampling_delay)


def configure_analog_in_channel(analog_in           : AnalogIn,
                                channel             : int,
                                channel_enable      : Optional[bool]=None,
                                channel_filter      : Optional[DwfAnalogInFilter]=None,
                                channel_range       : Optional[float]=None,
                                channel_offset      : Optional[float]=None,
                                channel_attenuation : Optional[float]=None,
                                channel_bandwidth   : Optional[float]=None
                               ) -> None:
    """Configure an |AnalogIn| instrument channel.

    Parameters:
        analog_in (AnalogIn): The |AnalogIn| instrument to be configured.
        channel (int): The analog input channel to be configured.
        channel_enable(Optional[bool]): If given, enable or disable the specified channel.
        channel_filter(Optional[DwfAnalogInFilter]): If given, configure channel's filter setting.
        channel_range(Optional[float]): If given, set the channel's range setting.
        channel_offset(Optional[float]): If given, set the channel's offset setting.
        channel_attenuation(Optional[float]): If given, set the channel's attenuation setting.
        channel_bandwidth(Optional[float]): If given, set the channel's bandwidth setting.

        Raises:
            DwfLibraryError: One of the settings could not be applied.
    """

    if channel_enable is not None:
        analog_in.channelEnableSet(channel, channel_enable)

    if channel_filter is not None:
        analog_in.channelFilterSet(channel, channel_filter)

    if channel_range is not None:
        analog_in.channelRangeSet(channel, channel_range)

    if channel_offset is not None:
        analog_in.channelOffsetSet(channel, channel_offset)

    if channel_attenuation is not None:
        analog_in.channelAttenuationSet(channel, channel_attenuation)

    if channel_bandwidth is not None:
        analog_in.channelBandwidthSet(channel, channel_bandwidth)

#######################################################################################################################
##                                                                                                                   ##
##                     Configuration support for the AnalogOut instrument and its channel nodes                      ##
##                                                                                                                   ##
#######################################################################################################################

def configure_analog_out_channel(analog_out           : AnalogOut,
                                 channel              : int,
                                 wait_duration        : Optional[float]=None,
                                 run_duration         : Optional[float]=None,
                                 repeat_trigger_flag  : Optional[bool]=None,
                                 repeat               : Optional[int]=None,
                                 master_channel_index : Optional[int]=None,
                                 trigger_source       : Optional[DwfTriggerSource]=None,
                                 trigger_slope        : Optional[DwfTriggerSlope]=None,
                                 mode                 : Optional[DwfAnalogOutMode]=None,
                                 idle                 : Optional[DwfAnalogOutIdle]=None,
                                 limitation           : Optional[float]=None,
                                 custom_am_fm_enable  : Optional[bool]=None
                                ) -> None:
    """Configure an analog output channel.

    Parameters:
        analog_out (AnalogOut)                     : The |AnalogOut| instrument to be configured.
        channel (int)                              : The analog output channel to be configured.
        wait_duration(Optional[float])             : If given, enable or disable the specified channel.
        run_duration(Optional[float])              : If given, configure channel's filter setting.
        repeat_trigger_flag(Optional[bool])        : If given, set the channel's repeat trigger flag.
        repeat(Optional[int])                      : If given, set the channel's repeat setting.
        master_channel_index(Optional[int])        : If given, set the channel's master channel setting.
        trigger_source(Optional[DwfTriggerSource]) : If given, set the channel's trigger source.
        trigger_slope(Optional[DwfTriggerSlope])   : If given, set the channel's trigger slope.
        mode(Optional[DwfAnalogOutMode])           : If given, set the channel's mode.
        idle(Optional[DwfAnalogOutIdle])           : If given, set the channel's idle behavior.
        limitation(Optional[float])                : If given, set the channel's limitation setting.
        custom_am_fm_enable(Optional[bool])        : If given, set the channel's custom AM/FM enable setting.

        Raises:
            DwfLibraryError: One of the settings could not be applied.
    """

    if wait_duration is not None:
        analog_out.waitSet(channel, wait_duration)

    if run_duration is not None:
        analog_out.runSet(channel, run_duration)

    if repeat_trigger_flag is not None:
        analog_out.repeatTriggerSet(channel, repeat_trigger_flag)

    if repeat is not None:
        analog_out.repeatSet(channel, repeat)

    if master_channel_index is not None:
        analog_out.masterSet(channel, master_channel_index)

    if trigger_source is not None:
        analog_out.triggerSourceSet(channel, trigger_source)

    if trigger_slope is not None:
        analog_out.triggerSlopeSet(channel, trigger_slope)

    if mode is not None:
        analog_out.modeSet(channel, mode)

    if idle is not None:
        analog_out.idleSet(channel, idle)

    if limitation is not None:
        analog_out.limitationSet(channel, limitation)

    if custom_am_fm_enable is not None:
        analog_out.customAMFMEnableSet(channel, custom_am_fm_enable)


def configure_analog_out_channel_node(
        analog_out     : AnalogOut,
        channel        : int,
        node           : DwfAnalogOutNode,
        node_enable    : Optional[bool]=None,
        node_function  : Optional[DwfAnalogOutFunction]=None,
        node_frequency : Optional[float]=None,
        node_amplitude : Optional[float]=None,
        node_offset    : Optional[float]=None,
        node_symmetry  : Optional[float]=None,
        node_phase     : Optional[float]=None,
        node_data      : Optional[np.ndarray]=None) -> None:
    """Configure an analog output channel node.

    Parameters:
        analog_out (AnalogOut)                        : The |AnalogOut| instrument to be configured.
        channel (int)                                 : The analog output channel to be configured.
        node (DwfAnalogOutNode)                       : The analog output node to be configured.
        node_enable(Optional[float])                  : If given, set the channel node's enable state.
        node_function(Optional[DwfAnalogOutFunction]) : If given, set the channel node's function.
        node_frequency(Optional[float])               : If given, set the channel node's frequency.
        node_amplitude(Optional[float])               : If given, set the channel node's amplitude.
        node_offset(Optional[float])                  : If given, set the channel node's offset.
        node_symmetry(Optional[float])                : If given, set the channel node's symmetry.
        node_phase(Optional[float])                   : If given, set the channel node's phase.
        node_data(Optional[float])                    : If given, set the channel node's data.

        Raises:
            DwfLibraryError: One of the settings could not be applied.
    """

    if node_enable is not None:
        analog_out.nodeEnableSet(channel, node, node_enable)

    if node_function is not None:
        analog_out.nodeFunctionSet(channel, node, node_function)

    if node_frequency is not None:
        analog_out.nodeFrequencySet(channel, node, node_frequency)

    if node_amplitude is not None:
        analog_out.nodeAmplitudeSet(channel, node, node_amplitude)

    if node_offset is not None:
        analog_out.nodeOffsetSet(channel, node, node_offset)

    if node_symmetry is not None:
        analog_out.nodeSymmetrySet(channel, node, node_symmetry)

    if node_phase is not None:
        analog_out.nodePhaseSet(channel, node, node_phase)

    if node_data is not None:
        analog_out.nodeDataSet(channel, node, node_data)



"""The |pydwf.core.dwf_library| module provides a single class: |DwfLibrary|."""

import sys
import ctypes

from pydwf.core.auxiliary.dwf_function_signatures import dwf_function_signatures, dwf_version as expected_dwf_version
from pydwf.core.auxiliary.exceptions import PyDwfError, DwfLibraryError
from pydwf.core.auxiliary.enum_types import DwfErrorCode, DwfDeviceParameter
from pydwf.core.auxiliary.constants import RESULT_SUCCESS
from pydwf.core.auxiliary.typespec_ctypes import typespec_ctypes
from pydwf.core.api.device_enumeration import DeviceEnumeration
from pydwf.core.api.device_control import DeviceControl
from pydwf.core.api.spectrum import Spectrum


class DwfLibrary:
    """The |DwfLibrary| class provides access to miscellaneous library functionality through the handful of methods
    it provides, and to |device enumeration:link|, |device control:link|, and |signal processing:link| functionality
    via its |deviceEnum:link|, |deviceControl:link|, and |spectrum:link| attributes.

    .. rubric:: |DwfLibrary| attributes

    Attributes:
        deviceEnum (DeviceEnumeration):
            Provides access to the |device enumeration:link| functionality.
        deviceControl (DeviceControl):
            Provides access to the |device control:link| functionality.
        spectrum (Spectrum):
            Provides access to the |signal processing:link| functionality.

    .. rubric:: |DwfLibrary| methods
    """

    def __init__(self, check_library_version: bool = False) -> None:
        """Initialize a |DwfLibrary| instance.

        A single |DwfLibrary| instance should be created by a user of |pydwf| to serve as an entry point to all
        |pydwf| functionality.

        When initializing a |DwfLibrary|, the shared library |libdwf| on top of which |pydwf| is built is loaded into
        memory using Python's standard |ctypes:link| module.

        A version check can be enabled to make sure that the shared library version corresponds exactly to the version
        that was used while developing and testing the current |pydwf| version (|libdwf-version|), and an exception is
        raised if a mismatch is detected. Enabling this flag is recommended for critical applications.

        After passing the version check (if enabled), the functions provided by the shared library are type-annotated.
        This means that calls into the shared library with incompatible parameter types will raise an exception.
        This mechanism helps to catch many bugs while using |pydwf|.

        As a last initialization step, the :py:attr:`deviceEnum`, :py:attr:`deviceControl`, and :py:attr:`spectrum`
        attributes are initialized. They can be used by a user program to access the |device enumeration:link|,
        |device control:link| functionality, and |signal processing:link| functionality.

        Parameters:
            check_library_version (bool): If True, the version number of the C library will be checked against the
                version of the C library from which the type information used by |pydwf| was derived. In case of a
                mismatch, an exception will be raised.

        Raises:
            PyDwfError: The version check could not be performed due to an unexpected low-level error while querying
                the shared library version, or a version mismatch was detected.
        """

        try:
            if sys.platform.startswith("win"):
                lib = ctypes.cdll.dwf
            elif sys.platform.startswith("darwin"):
                lib = ctypes.cdll.LoadLibrary("/Library/Frameworks/dwf.framework/dwf")
            else:
                lib = ctypes.cdll.LoadLibrary("libdwf.so")
        except OSError as exception:
            raise PyDwfError("Unable to open DWF library using ctypes.") from exception

        if check_library_version:
            # Note that the 'FDwfGetVersion()' function has not yet been type-annotated here.
            c_version = ctypes.create_string_buffer(32)
            result = lib.FDwfGetVersion(c_version)
            if result != RESULT_SUCCESS:
                raise PyDwfError("Unable to verify library version.")
            actual_dwf_version = c_version.value.decode()
            if actual_dwf_version != expected_dwf_version:
                raise PyDwfError(
                    "DWF library version mismatch: pydwf module expects {},"
                    " but actual library is version {}.".format(
                     expected_dwf_version, actual_dwf_version))

        self._annotate_function_signatures(lib)

        self._lib = lib

        # Initialize sub-API instances and assign them to attributes.

        self.deviceEnum = DeviceEnumeration(self)
        self.deviceControl = DeviceControl(self)
        self.spectrum = Spectrum(self)

    @property
    def lib(self) -> ctypes.CDLL:
        """Return the |ctypes:link| shared library instance used to access the DWF library.

        This property is provided for internal |pydwf| use.

        Returns:
            ctypes.CDLL: The |ctypes| library.

        :meta private:
        """
        return self._lib

    @staticmethod
    def _annotate_function_signatures(lib: ctypes.CDLL) -> None:
        """Add |ctypes:link| return type and parameter type annotations for all known functions in the DWF
        shared library.

        This method uses a list of DWF function signatures derived from the *dwf.h* C header file to perform
        the type annotation.

        Parameters:
            lib (ctypes.CDLL): The shared library whose functions will be type-annotated.

        This method is used exclusively by the :py:meth:`__init__` method.
        """

        function_signatures = dwf_function_signatures(typespec_ctypes)

        for (name, restype, argtypes, UNUSED_obsolete_flag) in function_signatures:
            argtypes = [argtype for (argname, argtype) in argtypes]
            try:
                func = getattr(lib, name)
                func.restype = restype
                func.argtypes = argtypes
            except AttributeError:
                # Do not annotate functions that are not present in the shared library.
                # This can happen, for example, when pydwf is used with an older version of the shared library.
                pass

    def exception(self) -> DwfLibraryError:
        """Return an exception describing the most recent error.

        This method is used by |pydwf| to generate a descriptive |DwfLibraryError:link| exception
        in case a DWF C library function has just failed.

        This method should not be called by |pydwf| users. It is for internal |pydwf| use only.

        Note that this method *returns* an exception instance; it doesn't *raise* it.

        Returns:
            DwfLibraryError: An exception describing the error reported by the last DWF library call.

        :meta private:
        """
        return DwfLibraryError(self.getLastError(), self.getLastErrorMsg())

    def getLastError(self) -> DwfErrorCode:
        """Retrieve the last error code in the calling process.

        The error code is cleared when other API functions are called and is only set when an API function
        fails during execution.

        Note:
            When using |pydwf| there is no need to call this method directly, since low-level errors reported by
            the C library are automatically converted to a |DwfLibraryError:link| exception, which includes
            both the error code and the corresponding message.

        Returns:
            DwfErrorCode: The DWF error code of last API call.

        Raises:
            DwfLibraryError: the last error code cannot be retrieved.
        """
        c_dwferc = typespec_ctypes.DwfErrorCode()
        result = self._lib.FDwfGetLastError(c_dwferc)
        if result != RESULT_SUCCESS:
            # If the 'FDwfGetLastError()' call itself fails, we cannot get a proper error code or message.
            raise DwfLibraryError(None, "FDwfGetLastError() failed.")
        dwferc = DwfErrorCode(c_dwferc.value)
        return dwferc

    def getLastErrorMsg(self) -> str:
        """Retrieve the last error message.

        The error message is cleared when other API functions are called and is only set when an API function
        fails during execution.

        Note:
            When using |pydwf| there is no need to call this method directly, since low-level errors reported by
            the C library are automatically converted to a |DwfLibraryError:link| exception, which includes
            both the error code and the corresponding message.

        Returns:
            str: The error message of the last API call.

            The string may consist of multiple messages, separated by a newline character,
            that describe the events leading to the error.

        Raises:
            DwfLibraryError: The last error message cannot be retrieved.
        """
        c_error_message = ctypes.create_string_buffer(512)
        result = self._lib.FDwfGetLastErrorMsg(c_error_message)
        if result != RESULT_SUCCESS:
            raise DwfLibraryError(None, "FDwfGetLastErrorMsg() failed.")
        error_message = c_error_message.value.decode()
        return error_message

    def getVersion(self) -> str:
        """Retrieve the library version string.

        Returns:
            str: The version of the DWF C library, composed of major, minor, and build numbers
            (e.g., "|libdwf-version|").

        Raises:
            DwfLibraryError: The library version string cannot be retrieved.
        """
        c_version = ctypes.create_string_buffer(32)
        result = self._lib.FDwfGetVersion(c_version)
        if result != RESULT_SUCCESS:
            raise self.exception()
        version = c_version.value.decode()
        return version

    def paramSet(self, device_parameter: DwfDeviceParameter, value: int) -> None:
        """Configure a default device parameter value.

        Device parameters are settings of a specific |DwfDevice|.
        Refer to the |device parameters:link| section for more information.

        This method sets a default device parameter value to be used for devices that are
        opened subsequently.

        See Also:
            To set the parameter value of a specific |DwfDevice|, use the
            :py:meth:`DwfDevice.paramSet <pydwf.core.DwfDevice.paramSet>` method.

        Warning:
            The device parameter values are not checked to make sure they correspond to a valid
            value for the specific device parameter.

        Parameters:
            device_parameter (DwfDeviceParameter): The device parameter for which to set the default value.
            value (int): The default device parameter value.

        Raises:
            DwfLibraryError: The device parameter value cannot be set.
        """
        result = self._lib.FDwfParamSet(device_parameter.value, value)
        if result != RESULT_SUCCESS:
            raise self.exception()

    def paramGet(self, device_parameter: DwfDeviceParameter) -> int:
        """Return a default device parameter value.

        Device parameters are settings of a specific |DwfDevice|.
        Refer to the |device parameters:link| section for more information.

        This method retrieves device parameter values at the library level (i.e., not tied to a specific device).
        They are used as default device parameter values for devices that are opened subsequently.

        See Also:
            To get the parameter value of a specific |DwfDevice|, use the
            :py:meth:`DwfDevice.paramGet <pydwf.core.DwfDevice.paramGet>` method.

        Parameters:
            device_parameter (DwfParameter): The device parameter for which to get the value.

        Returns:
            int: The retrieved device parameter value.

        Raises:
            DwfLibraryError: The device parameter value cannot be retrieved.
        """
        c_value = typespec_ctypes.c_int()
        result = self._lib.FDwfParamGet(device_parameter.value, c_value)
        if result != RESULT_SUCCESS:
            raise self.exception()
        value = c_value.value
        return value



"""The |pydwf.core.api.protocol_spi| module provides a single class: |ProtocolSPI|."""

# pylint: disable=too-many-lines

from typing import List

from pydwf.core.dwf_device_subapi import AbstractDwfDeviceSubAPI

from pydwf.core.auxiliary.typespec_ctypes import typespec_ctypes
from pydwf.core.auxiliary.constants import RESULT_SUCCESS
from pydwf.core.auxiliary.enum_types import DwfDigitalOutIdle


class ProtocolSPI(AbstractDwfDeviceSubAPI):

    """The |ProtocolSPI| class provides access to the SPI protocol functionality of a |DwfDevice:link|.

    Attention:
        Users of |pydwf| should not create instances of this class directly.

        It is instantiated during initialization of a |DwfDevice| and subsequently accessible via its
        |protocol.spi:link| attribute.
    """

    # pylint: disable=too-many-public-methods

    def reset(self) -> None:
        """Reset the SPI protocol support.

        Raises:
            DwfLibraryError: An error occurred while executing the *reset* operation.
        """
        result = self.lib.FDwfDigitalSpiReset(self.hdwf)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def frequencySet(self, frequency: float) -> None:
        """Set the SPI frequency, in Hz.

        Parameters:
            frequency (float): SPI frequency, in Hz.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalSpiFrequencySet(self.hdwf, frequency)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def clockSet(self, channel_index: int) -> None:
        """Set the digital channel (pin) for the SPI clock signal.

        Parameters:
            channel_index (int):
                The digital channel (pin) where the SPI clock signal will be generated.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalSpiClockSet(self.hdwf, channel_index)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def dataSet(self, spi_data_bit: int, channel_index: int) -> None:
        """Set the digital channel (pin) for an SPI data bit.

        Parameters:
            spi_data_bit (int):

                The data bit to configure:

                * 0 — DQ0 / MOSI / SISO
                * 1 — DQ1 / MISO
                * 2 — DQ2
                * 3 — DQ3

            channel_index (int):
                The digital channel (pin) for this data bit.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalSpiDataSet(self.hdwf, spi_data_bit, channel_index)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def idleSet(self, spi_data_bit: int, idle_mode: DwfDigitalOutIdle) -> None:
        """Set the idle behavior for an SPI data bit.

        Parameters:
            spi_data_bit (int):

                The data bit to configure:

                * 0 — DQ0 / MOSI / SISO
                * 1 — DQ1 / MISO
                * 2 — DQ2
                * 3 — DQ3

            idle_mode (DwfDigitalOutIdle):
                The idle behavior of this bit.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """

        result = self.lib.FDwfDigitalSpiIdleSet(self.hdwf, spi_data_bit, idle_mode.value)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def modeSet(self, spi_mode: int) -> None:
        """Set the SPI mode.

        Parameters:
            spi_mode (int):

                The values for CPOL (polarity) and CPHA (phase) to use with the attached slave device:

                * 0 — CPOL = 0, CPHA = 0
                * 1 — CPOL = 0, CPHA = 1
                * 2 — CPOL = 1, CPHA = 0
                * 3 — CPOL = 1, CPHA = 1

                Refer to the slave device's datasheet to select the correct value.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalSpiModeSet(self.hdwf, spi_mode)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def orderSet(self, bit_order: int) -> None:
        """Set the SPI data bit order.

        Parameters:
            bit_order (int):

                Select the bit order of each word sent out:

                * 1 — MSB first, LSB last
                * 0 — LSB first, MSB last

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalSpiOrderSet(self.hdwf, bit_order)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def delaySet(self, start: int, cmd: int, word: int, stop: int) -> None:
        """Set the SPI delays.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalSpiDelaySet(self.hdwf, start, cmd, word, stop)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def selectSet(self, channel: int, level: int) -> None:
        """Set SPI device select channel and level.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalSpiSelectSet(self.hdwf, channel, level)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def select(self, channel_index: int, level: int) -> None:
        """Set the chip select (CS) status.

        Parameters:

            channel_index (int):
                The digital channel (pin) for the Chip Select signal.

            level (int):

                The Chip Select level to configure.

                *  0 — low
                *  1 — high
                * -1 — Z (high impedance)

                The CS (chip select) is an active-low signal, from the SPI bus master to a
                specific SPI slave device. Before starting a bus request, the master should
                set CS to 0 for the chip it wants to talk to.

                Each slave on an SPI bus has its own CS line. At most one of them should be
                selected at any time.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalSpiSelect(self.hdwf, channel_index, level)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def writeRead(self, transfer_type: int, bits_per_word: int, tx: List[int]) -> List[int]:
        """Write and read multiple SPI data-words, with up to 8 bits per data-word.

        Parameters:
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…8).
            tx (List[int]):
                The data-words to write.

        Returns:
            List[int]: The data-words received.

        Raises:
            DwfLibraryError: An error occurred while executing the write/read operation.
        """
        tx_list = list(tx)

        number_of_words = len(tx_list)

        buffer_type = typespec_ctypes.c_unsigned_char * number_of_words

        tx_buffer = buffer_type(*tx_list)
        rx_buffer = buffer_type()

        result = self.lib.FDwfDigitalSpiWriteRead(
            self.hdwf,
            transfer_type,
            bits_per_word,
            tx_buffer,
            number_of_words,
            rx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        rx_list = list(rx_buffer)

        return rx_list

    def writeRead16(self, transfer_type: int, bits_per_word: int, tx: List[int]) -> List[int]:
        """Write and read multiple SPI data-words, with up to 16 bits per data-word.

        Parameters:
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…16).
            tx (list[int]):
                The data-words to write.

        Returns:
            List[int]: The data-words received.

        Raises:
            DwfLibraryError: An error occurred while executing the write/read operation.
        """
        tx_list = list(tx)

        number_of_words = len(tx_list)

        buffer_type = typespec_ctypes.c_unsigned_short * number_of_words

        tx_buffer = buffer_type(*tx_list)
        rx_buffer = buffer_type()

        result = self.lib.FDwfDigitalSpiWriteRead16(
            self.hdwf,
            transfer_type,
            bits_per_word,
            tx_buffer,
            number_of_words,
            rx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        rx_list = list(rx_buffer)

        return rx_list

    def writeRead32(self, transfer_type: int, bits_per_word: int, tx: List[int]) -> List[int]:
        """Write and read multiple SPI data-words, with up to 32 bits per data-word.

        Parameters:
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…32).
            tx (List[int]):
                The data-words to write.

        Returns:
            List[int]: The data-words received.

        Raises:
            DwfLibraryError: An error occurred while executing the write/read operation.
        """

        tx_list = list(tx)

        number_of_words = len(tx_list)

        buffer_type = typespec_ctypes.c_unsigned_int * number_of_words

        tx_buffer = buffer_type(*tx_list)
        rx_buffer = buffer_type()

        result = self.lib.FDwfDigitalSpiWriteRead32(
            self.hdwf,
            transfer_type,
            bits_per_word,
            tx_buffer,
            number_of_words,
            rx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        rx_list = list(rx_buffer)

        return rx_list

    def read(self, transfer_type: int, bits_per_word: int, number_of_words: int) -> List[int]:
        """Read multiple SPI data-words, with up to 8 bits per data-word.

        Parameters:
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…8).
            number_of_words (int):
                The number of data-words to read.

        Returns:
            List[int]: The data-words received.

        Raises:
            DwfLibraryError: An error occurred while executing the read operation.
        """

        buffer_type = typespec_ctypes.c_unsigned_char * number_of_words

        rx_buffer = buffer_type()

        result = self.lib.FDwfDigitalSpiRead(
            self.hdwf,
            transfer_type,
            bits_per_word,
            rx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        rx_list = list(rx_buffer)

        return rx_list

    def readOne(self, transfer_type: int, bits_per_word: int) -> int:
        """Read a single SPI data-word, with up to 32 bits.

        Parameters:
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits of the data-word (1…32).

        Returns:
            int: The data-word received.

        Raises:
            DwfLibraryError: An error occurred while executing the read operation.
        """
        c_rx = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalSpiReadOne(self.hdwf, transfer_type, bits_per_word, c_rx)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        rx = c_rx.value
        return rx

    def read16(self, transfer_type: int, bits_per_word: int, number_of_words: int) -> List[int]:
        """Read multiple SPI data-words, with up to 16 bits per data-word.

        Parameters:
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…16).
            number_of_words (int):
                The number of data-words to read.

        Returns:
            List[int]: The data-words received.

        Raises:
            DwfLibraryError: An error occurred while executing the read operation.
        """

        buffer_type = typespec_ctypes.c_unsigned_short * number_of_words

        rx_buffer = buffer_type()

        result = self.lib.FDwfDigitalSpiRead16(
            self.hdwf,
            transfer_type,
            bits_per_word,
            rx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        rx_list = list(rx_buffer)

        return rx_list

    def read32(self, transfer_type: int, bits_per_word: int, number_of_words: int) -> List[int]:
        """Read multiple SPI data-words, with up to 32 bits per data-word.

        Parameters:
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…32).
            number_of_words (int):
                The number of data-words to read.

        Returns:
            List[int]: The data-words received.

        Raises:
            DwfLibraryError: An error occurred while executing the read operation.
        """

        buffer_type = typespec_ctypes.c_unsigned_int * number_of_words

        rx_buffer = buffer_type()

        result = self.lib.FDwfDigitalSpiRead32(
            self.hdwf,
            transfer_type,
            bits_per_word,
            rx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        rx_list = list(rx_buffer)

        return rx_list

    def write(self, transfer_type: int, bits_per_word: int, tx: List[int]) -> None:
        """Write multiple SPI data-words, with up to 32 bits per data-word.

        Parameters:
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…32).
            tx (List[int]):
                The data-words to write.

        Raises:
            DwfLibraryError: An error occurred while executing the write operation.
        """
        tx_list = list(tx)

        number_of_words = len(tx_list)

        buffer_type = typespec_ctypes.c_unsigned_char * number_of_words

        tx_buffer = buffer_type(*tx_list)

        result = self.lib.FDwfDigitalSpiWrite(
            self.hdwf,
            transfer_type,
            bits_per_word,
            tx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def writeOne(self, transfer_type: int, bits_per_word: int, tx: int) -> None:
        """Write a single SPI data-word, with up to 32 bits.

        Parameters:
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits of the data-word (1…32).
            tx (int):
                The data-word to write.

        Raises:
            DwfLibraryError: An error occurred while executing the write operation.
        """
        result = self.lib.FDwfDigitalSpiWriteOne(self.hdwf, transfer_type, bits_per_word, tx)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def write16(self, transfer_type: int, bits_per_word: int, tx: List[int]) -> None:
        """Write multiple SPI data-words, with up to 16 bits per data-word.

        Parameters:
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…16).
            tx (List[int]):
                The data-words to write.

        Raises:
            DwfLibraryError: An error occurred while executing the write operation.
        """
        tx_list = list(tx)

        number_of_words = len(tx_list)

        buffer_type = typespec_ctypes.c_unsigned_short * number_of_words

        tx_buffer = buffer_type(*tx_list)

        result = self.lib.FDwfDigitalSpiWrite16(
            self.hdwf,
            transfer_type,
            bits_per_word,
            tx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def write32(self, transfer_type: int, bits_per_word: int, tx: List[int]) -> None:
        """Write multiple SPI data-words, with up to 32 bits per data-word.

        Parameters:
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…32).
            tx (List[int]):
                The data-words to write.

        Raises:
            DwfLibraryError: An error occurred while executing the write operation.
        """
        tx_list = list(tx)

        number_of_words = len(tx_list)

        buffer_type = typespec_ctypes.c_unsigned_int * number_of_words

        tx_buffer = buffer_type(*tx_list)

        result = self.lib.FDwfDigitalSpiWrite32(
            self.hdwf,
            transfer_type,
            bits_per_word,
            tx_buffer, number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def cmdWriteRead(self, command_bits: int, command_value: int, dummy_bits: int,
                     transfer_type: int, bits_per_word: int, tx: List[int]) -> List[int]:
        """Send command and write and read multiple SPI data-words, with up to 8 bits per data-word.

        Parameters:
            command_bits (int):
                The number of command bits.
            command_value (int):
                The command value.
            dummy_bits (int):
                The number of dummy bits before the data transfer.
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…8).
            tx (List[int]):
                The data-words to write.

        Returns:
            List[int]: The data-words received.

        Raises:
            DwfLibraryError: An error occurred while executing the write/read operation.
        """

        tx_list = list(tx)

        number_of_words = len(tx_list)

        buffer_type = typespec_ctypes.c_unsigned_char * number_of_words

        tx_buffer = buffer_type(*tx_list)
        rx_buffer = buffer_type()

        result = self.lib.FDwfDigitalSpiCmdWriteRead(
            self.hdwf,
            command_bits,
            command_value,
            dummy_bits,
            transfer_type,
            bits_per_word,
            tx_buffer,
            number_of_words,
            rx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        rx_list = list(rx_buffer)

        return rx_list

    def cmdWriteRead16(self, command_bits: int, command_value: int, dummy_bits: int,
                       transfer_type: int, bits_per_word: int, tx: List[int]) -> List[int]:
        """Send command and write and read multiple SPI data-words, with up to 16 bits per data-word.

        Parameters:
            command_bits (int):
                The number of command bits.
            command_value (int):
                The command value.
            dummy_bits (int):
                The number of dummy bits before the data transfer.
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…16).
            tx (list[int]):
                The data-words to write.

        Returns:
            List[int]: The data-words received.

        Raises:
            DwfLibraryError: An error occurred while executing the write/read operation.
        """

        tx_list = list(tx)

        number_of_words = len(tx_list)

        buffer_type = typespec_ctypes.c_unsigned_short * number_of_words

        tx_buffer = buffer_type(*tx_list)
        rx_buffer = buffer_type()

        result = self.lib.FDwfDigitalSpiCmdWriteRead16(
            self.hdwf,
            command_bits,
            command_value,
            dummy_bits,
            transfer_type,
            bits_per_word,
            tx_buffer,
            number_of_words,
            rx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        rx_list = list(rx_buffer)

        return rx_list

    def cmdWriteRead32(self, command_bits: int, command_value: int, dummy_bits: int,
                       transfer_type: int, bits_per_word: int, tx: List[int]) -> List[int]:
        """Send command and write and read multiple SPI data-words, with up to 32 bits per data-word.

        Parameters:
            command_bits (int):
                The number of command bits.
            command_value (int):
                The command value.
            dummy_bits (int):
                The number of dummy bits before the data transfer.
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…32).
            tx (List[int]):
                The data-words to write.

        Returns:
            List[int]: The data-words received.

        Raises:
            DwfLibraryError: An error occurred while executing the write/read operation.
        """

        tx_list = list(tx)

        number_of_words = len(tx_list)

        buffer_type = typespec_ctypes.c_unsigned_int * number_of_words

        tx_buffer = buffer_type(*tx_list)
        rx_buffer = buffer_type()

        result = self.lib.FDwfDigitalSpiCmdWriteRead32(
            self.hdwf,
            command_bits,
            command_value,
            dummy_bits,
            transfer_type,
            bits_per_word,
            tx_buffer,
            number_of_words,
            rx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        rx_list = list(rx_buffer)

        return rx_list

    def cmdRead(self, command_bits: int, command_value: int, dummy_bits: int,
                transfer_type: int, bits_per_word: int, number_of_words: int) -> List[int]:
        """Send command and read multiple SPI data-words, with up to 8 bits per data-word.

        Parameters:
            command_bits (int):
                The number of command bits.
            command_value (int):
                The command value.
            dummy_bits (int):
                The number of dummy bits before the data transfer.
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…8).
            number_of_words (int):
                The number of data-words to read.

        Returns:
            List[int]: The data-words received.

        Raises:
            DwfLibraryError: An error occurred while executing the read operation.
        """

        buffer_type = typespec_ctypes.c_unsigned_char * number_of_words

        rx_buffer = buffer_type()

        result = self.lib.FDwfDigitalSpiCmdRead(
            self.hdwf,
            command_bits,
            command_value,
            dummy_bits,
            transfer_type,
            bits_per_word,
            rx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        rx_list = list(rx_buffer)

        return rx_list

    def cmReadOne(self, command_bits: int, command_value: int, dummy_bits: int,
                  transfer_type: int, bits_per_word: int) -> int:
        """Send command and read a single SPI data-word, with up to 32 bits.

        Parameters:
            command_bits (int):
                The number of command bits.
            command_value (int):
                The command value.
            dummy_bits (int):
                The number of dummy bits before the data transfer.
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits of the data-word (1…32).

        Returns:
            int: The data-word received.

        Raises:
            DwfLibraryError: An error occurred while executing the read operation.
        """
        c_rx = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalSpiCmdReadOne(
            self.hdwf,
            command_bits,
            command_value,
            dummy_bits,
            transfer_type,
            bits_per_word,
            c_rx)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        rx = c_rx.value
        return rx

    def cmdRead16(self, command_bits: int, command_value: int, dummy_bits: int,
                  transfer_type: int, bits_per_word: int, number_of_words: int) -> List[int]:
        """Send command and read multiple SPI data-words, with up to 16 bits per data-word.

        Parameters:
            command_bits (int):
                The number of command bits.
            command_value (int):
                The command value.
            dummy_bits (int):
                The number of dummy bits before the data transfer.
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…16).
            number_of_words (int):
                The number of data-words to read.

        Returns:
            List[int]: The data-words received.

        Raises:
            DwfLibraryError: An error occurred while executing the read operation.
        """

        buffer_type = typespec_ctypes.c_unsigned_short * number_of_words

        rx_buffer = buffer_type()

        result = self.lib.FDwfDigitalSpiCmdRead16(
            self.hdwf,
            command_bits,
            command_value,
            dummy_bits,
            transfer_type,
            bits_per_word,
            rx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        rx_list = list(rx_buffer)

        return rx_list

    def cmdRead32(self, command_bits: int, command_value: int, dummy_bits: int,
                  transfer_type: int, bits_per_word: int, number_of_words: int) -> List[int]:
        """Send command and read multiple SPI data-words, with up to 32 bits per data-word.

        Parameters:
            command_bits (int):
                The number of command bits.
            command_value (int):
                The command value.
            dummy_bits (int):
                The number of dummy bits before the data transfer.
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…32).
            number_of_words (int):
                The number of data-words to read.

        Returns:
            List[int]: The data-words received.

        Raises:
            DwfLibraryError: An error occurred while executing the read operation.
        """

        buffer_type = typespec_ctypes.c_unsigned_int * number_of_words

        rx_buffer = buffer_type()

        result = self.lib.FDwfDigitalSpiCmdRead32(
            self.hdwf,
            command_bits,
            command_value,
            dummy_bits,
            transfer_type,
            bits_per_word,
            rx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        rx_list = list(rx_buffer)

        return rx_list

    def cmdWrite(self, command_bits: int, command_value: int, dummy_bits: int,
                 transfer_type: int, bits_per_word: int, tx: List[int]) -> None:
        """Send command and write multiple SPI data-words, with up to 32 bits per data-word.

        Parameters:
            command_bits (int):
                The number of command bits.
            command_value (int):
                The command value.
            dummy_bits (int):
                The number of dummy bits before the data transfer.
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…32).
            tx (List[int]):
                The data-words to write.

        Raises:
            DwfLibraryError: An error occurred while executing the write operation.
        """
        tx_list = list(tx)

        number_of_words = len(tx_list)

        buffer_type = typespec_ctypes.c_unsigned_char * number_of_words

        tx_buffer = buffer_type(*tx_list)

        result = self.lib.FDwfDigitalSpiCmdWrite(
            self.hdwf,
            command_bits,
            command_value,
            dummy_bits,
            transfer_type,
            bits_per_word,
            tx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def cmdWriteOne(self, command_bits: int, command_value: int, dummy_bits: int,
                    transfer_type: int, bits_per_word: int, tx: int) -> None:
        """Send command and write a single SPI data-word, with up to 32 bits.

        Parameters:
            command_bits (int):
                The number of command bits.
            command_value (int):
                The command value.
            dummy_bits (int):
                The number of dummy bits before the data transfer.
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits of the data-word (1…32).
            tx (int):
                The data-word to write.

        Raises:
            DwfLibraryError: An error occurred while executing the write operation.
        """
        result = self.lib.FDwfDigitalSpiCmdWriteOne(
            self.hdwf,
            command_bits,
            command_value,
            dummy_bits,
            transfer_type,
            bits_per_word,
            tx)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def cmdWrite16(self, command_bits: int, command_value: int, dummy_bits: int,
                   transfer_type: int, bits_per_word: int, tx: List[int]) -> None:
        """Send command and write multiple SPI data-words, with up to 16 bits per data-word.

        Parameters:
            command_bits (int):
                The number of command bits.
            command_value (int):
                The command value.
            dummy_bits (int):
                The number of dummy bits before the data transfer.
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…16).
            tx (List[int]):
                The data-words to write.

        Raises:
            DwfLibraryError: An error occurred while executing the write operation.
        """
        tx_list = list(tx)

        number_of_words = len(tx_list)

        buffer_type = typespec_ctypes.c_unsigned_short * number_of_words

        tx_buffer = buffer_type(*tx_list)

        result = self.lib.FDwfDigitalSpiCmdWrite16(
            self.hdwf,
            command_bits,
            command_value,
            dummy_bits,
            transfer_type,
            bits_per_word,
            tx_buffer,
            number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def cmdWrite32(self, command_bits: int, command_value: int, dummy_bits: int,
                   transfer_type: int, bits_per_word: int, tx: List[int]) -> None:
        """Send command and write multiple SPI data-words, with up to 32 bits per data-word.

        Parameters:
            command_bits (int):
                The number of command bits.
            command_value (int):
                The command value.
            dummy_bits (int):
                The number of dummy bits before the data transfer.
            transfer_type (int):
                * 0 — SISO
                * 1 — MOSI/MISO
                * 2 — dual
                * 4 — quad
            bits_per_word (int):
                The number of bits per data-word (1…32).
            tx (List[int]):
                The data-words to write.

        Raises:
            DwfLibraryError: An error occurred while executing the write operation.
        """
        tx_list = list(tx)

        number_of_words = len(tx_list)

        buffer_type = typespec_ctypes.c_unsigned_int * number_of_words

        tx_buffer = buffer_type(*tx_list)

        result = self.lib.FDwfDigitalSpiCmdWrite32(
            self.hdwf,
            command_bits,
            command_value,
            dummy_bits,
            transfer_type,
            bits_per_word,
            tx_buffer, number_of_words)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()



"""The |pydwf.core.api.ProtocolI2C| module provides a single class: |ProtocolI2C|."""

from typing import List, Tuple

from pydwf.core.dwf_device_subapi import AbstractDwfDeviceSubAPI

from pydwf.core.auxiliary.typespec_ctypes import typespec_ctypes
from pydwf.core.auxiliary.constants import RESULT_SUCCESS


class ProtocolI2C(AbstractDwfDeviceSubAPI):
    """The |ProtocolI2C| class provides access to the I²C protocol functionality of a |DwfDevice:link|.

    Attention:
        Users of |pydwf| should not create instances of this class directly.

        It is instantiated during initialization of a |DwfDevice| and subsequently accessible via its
        |protocol.i2c:link| attribute.
    """

    def reset(self) -> None:
        """Reset the I²C protocol instrument.

        Raises:
            DwfLibraryError: An error occurred while executing the *reset* operation.
        """
        result = self.lib.FDwfDigitalI2cReset(self.hdwf)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def clear(self) -> bool:
        """Clear the I²C bus.

        Detect and try to solve an I²C bus lockup condition.

        Todo:
            The precise behavior of this method needs to be understood and documented.

        Returns:
            bool: True if the bus is clear after the operation, False otherwise.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_bus_free = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalI2cClear(self.hdwf, c_bus_free)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        bus_free = bool(c_bus_free.value)
        return bus_free

    def stretchSet(self, stretch_enable: int) -> None:
        """Set I²C stretch behavior.

        Todo:
            The precise behavior of this method needs to be understood and documented.

        Parameters:
            stretch_enable (bool): True to enable stretch, False to disable.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalI2cStretchSet(self.hdwf, bool(stretch_enable))
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def rateSet(self, data_rate: float) -> None:
        """Set the I²C data rate, in Hz.

        Parameters:
            data_rate (float): I²C data rate. Often-encountered rates are 100 kHz and 400 kHz, but many
                modern I²C devices support higher data rates. Check the datasheet of your device.

                The default value is 100 kHz.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalI2cRateSet(self.hdwf, data_rate)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def timeoutSet(self, timeout: float) -> None:
        """Set the I²C timeout, in seconds.

        Parameters:
            timeout (float): I²C timeout. The default value is 1 second.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalI2cTimeoutSet(self.hdwf, timeout)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def readNakSet(self, nak_last_read_byte: int) -> None:
        """Set read NAK state.

        Todo:
            The precise behavior of this method needs to be understood and documented.

        Parameters:
            nak_last_read_byte (int): (undocumented)

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalI2cReadNakSet(self.hdwf, nak_last_read_byte)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def sclSet(self, channel_index: int) -> None:
        """Set the digital channel (pin) where the I²C clock (SCL) signal is transmitted.

        Parameters:
            channel_index (int): The digital channel (pin) on which to generate the SCL clock.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalI2cSclSet(self.hdwf, channel_index)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def sdaSet(self, channel_index: int) -> None:
        """Set the digital channel (pin) where the I²C data (SDA) signal is transmitted/received.

        Parameters:
            channel_index (int): The digital channel (pin) on which to send/receive SDA data.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalI2cSdaSet(self.hdwf, channel_index)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def writeRead(self, address: int, tx: List[int], number_of_rx_bytes: int) -> Tuple[int, List[int]]:
        """Perform a combined I²C write/read operation.

        Parameters:
            address (int): The I²C address of the target device.
            tx (List[int]): The octets to send.
            number_of_rx_bytes (int): The number of octets to receive.

        Returns:
            Tuple[int, List[int]]:
                The first element is the NAK indication; the second element is a list of octet values received.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_nak = typespec_ctypes.c_int()

        tx_list = list(tx)

        number_of_tx_bytes = len(tx_list)

        tx_buffer_type = typespec_ctypes.c_unsigned_char * number_of_tx_bytes
        rx_buffer_type = typespec_ctypes.c_unsigned_char * number_of_rx_bytes

        tx_buffer = tx_buffer_type(*tx_list)
        rx_buffer = rx_buffer_type()

        result = self.lib.FDwfDigitalI2cWriteRead(
            self.hdwf,
            address,
            tx_buffer,
            number_of_tx_bytes,
            rx_buffer,
            number_of_rx_bytes,
            c_nak)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        nak = c_nak.value

        rx_list = list(rx_buffer)

        return (nak, rx_list)

    def read(self, address: int, number_of_words: int) -> Tuple[int, List[int]]:
        """Perform an I²C read operation.

        Parameters:
            address (int): The I²C address of the target device.
            number_of_words (int): The number of octets to receive.

        Returns:
            Tuple[int, List[int]]:
                The first element is the NAK indication; the second element is a list of octet values received.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """

        c_nak = typespec_ctypes.c_int()

        buffer_type = typespec_ctypes.c_unsigned_char * number_of_words

        rx_buffer = buffer_type()

        result = self.lib.FDwfDigitalI2cRead(self.hdwf, address, rx_buffer, number_of_words, c_nak)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        nak = c_nak.value

        rx_list = list(rx_buffer)

        return (nak, rx_list)

    def write(self, address: int, tx: List[int]) -> int:
        """Perform an I²C write operation.

        Parameters:
            address (int): The I²C address of the target device.
            tx (List[int]): The octets to send.

        Returns:
            int: The NAK indication.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """

        c_nak = typespec_ctypes.c_int()

        tx_list = list(tx)

        number_of_words = len(tx_list)

        buffer_type = typespec_ctypes.c_unsigned_char * number_of_words

        tx_buffer = buffer_type(*tx_list)

        result = self.lib.FDwfDigitalI2cWrite(self.hdwf, address, tx_buffer, number_of_words, c_nak)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        nak = c_nak.value

        return nak

    def writeOne(self, address: int, tx: int) -> int:
        """Perform an I²C write operation of a single octet.

        Parameters:
            address (int): The I²C address of the target device.
            tx (int): The single octet to send.

        Returns:
            int: The NAK indication.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_nak = typespec_ctypes.c_int()

        result = self.lib.FDwfDigitalI2cWriteOne(self.hdwf, address, tx, c_nak)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        nak = c_nak.value

        return nak

    def spyStart(self) -> None:
        """Start I²C spy.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalI2cSpyStart(self.hdwf)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def spyStatus(self, max_data_size: int) -> Tuple[int, int, List[int], int]:
        """Get I²C spy status.

        Returns:
            Tuple[int, int, List[int], int]: A tuple (start, stop, data-values, nak-indicator).

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """

        c_nak   = typespec_ctypes.c_int()
        c_start = typespec_ctypes.c_int()
        c_stop  = typespec_ctypes.c_int()

        buffer_type = typespec_ctypes.c_unsigned_char * max_data_size
        c_data = buffer_type()

        c_data_size = typespec_ctypes.c_int(max_data_size)

        result = self.lib.FDwfDigitalI2cSpyStatus(self.hdwf, c_start, c_stop, c_data, c_data_size, c_nak)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        start = c_start.value
        stop  = c_stop.value
        nak   = c_nak.value

        data_list = list(c_data)[:c_data_size.value]  # Only return the first 'c_data_size' words.

        return (start, stop, data_list, nak)



"""The |pydwf.utilities| package provides utility functions for managing Digilent Waveforms devices.

It provides high-level functions that reflect best-practice implementations for common use-cases of |pydwf|.

Currently, it only provides a single function, but several more are expected in the future.
"""

# Make the openDwfDevice function directly available from the pydwf.utilities package.
from pydwf.utilities.open_dwf_device import openDwfDevice



"""The |pydwf.core.auxiliary.exceptions| module provides the |PyDwfError| and |DwfLibraryError|
exception classes."""

from typing import Optional

from pydwf.core.auxiliary.enum_types import DwfErrorCode


class PyDwfError(Exception):
    """A |PyDwfError| exception represents any error in |pydwf| (caused by the underlying C API or otherwise).

    It is a trivial (empty) specialization of the built-in |Exception| class.
    """


class DwfLibraryError(PyDwfError):
    """A |DwfLibraryError| exception represents an error reported by one of the DWF C library functions.

    This class derives from |PyDwfError|, making it easier for scripts to catch any exception originating in |pydwf|.

    The following attributes are provided:

    Attributes:

        code (Optional[DwfErrorCode]):
            DWF error code as reported by the C library, if available.

        msg (Optional[str]):
            DWF error message as reported by the C library, if available.
            It may contain multiple single-line messages, separated by a newline character.
    """
    def __init__(self, code: Optional[DwfErrorCode], msg: Optional[str]) -> None:
        super().__init__(self)
        self.code = code
        self.msg = msg

    def __str__(self) -> str:
        if self.code is None:
            error_string = "DWF Library Error (unspecified)"
        else:
            error_string = "DWF Library Error {!r} ({})".format(self.code.name, self.code.value)

        if self.msg is not None:
            error_string = "{}: {!r}".format(error_string, self.msg.strip())

        return error_string



"""This module defines symbolic constants that are used by the C API."""

HDWF_NONE = 0
"""A HDWF value representing a bad device handle."""

RESULT_SUCCESS = 1
"""This value is returned by all C API functions in case of success.
The return type of API functions used to be a boolean, with true indicating success.
"""



"""The |pydwf.core.api.protocol_can| module provides a single class: |ProtocolCAN|."""

from typing import Tuple
import ctypes

from pydwf.core.dwf_device_subapi import AbstractDwfDeviceSubAPI

from pydwf.core.auxiliary.typespec_ctypes import typespec_ctypes
from pydwf.core.auxiliary.constants import RESULT_SUCCESS
from pydwf.core.auxiliary.exceptions import PyDwfError


class ProtocolCAN(AbstractDwfDeviceSubAPI):
    """The |ProtocolCAN| class provides access to the CAN bus protocol functionality of a |DwfDevice:link|.

    Attention:
        Users of |pydwf| should not create instances of this class directly.

        It is instantiated during initialization of a |DwfDevice| and subsequently accessible via its
        |protocol.can:link| attribute.
    """

    def reset(self) -> None:
        """Reset the CAN bus protocol functionality.

        Raises:
            DwfLibraryError: An error occurred while executing the *reset* operation.
        """
        result = self.lib.FDwfDigitalCanReset(self.hdwf)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def rateSet(self, data_rate: float) -> None:
        """Set the CAN bus data rate, in Hz.

        Parameters:
            data_rate (float): The data-rate used by the receiver and transmitter.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalCanRateSet(self.hdwf, data_rate)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def polaritySet(self, polarity: bool) -> None:
        """Set the CAN bus polarity.

        Parameters:
            polarity (bool): If True, set polarity to *high*.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalCanPolaritySet(self.hdwf, int(polarity))
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def txSet(self, channel_index: int) -> None:
        """Set the digital channel (pin) where the outgoing (TX) signal will be transmitted.

        Parameters:
            channel_index (int): The digital channel (pin) on which to transmit data.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalCanTxSet(self.hdwf, channel_index)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def rxSet(self, channel_index: int) -> None:
        """Set the digital channel (pin) where the incoming (RX) signal is received.

        Parameters:
            channel_index (int): The digital channel (pin) on which to receive data.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalCanRxSet(self.hdwf, channel_index)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def tx(self, v_id: int, extended: bool, remote: bool, data: bytes) -> None:
        """Transmit outgoing CAN bus data.

        Parameters:
            v_id (int): (to be documented).
            extended (bool): (to be documented).
            remote (bool): (to be documented).
            data (bytes): The data to be transmitted. Should be at most 8 bytes.

        Raises:
            PyDwfError: A request to transmit more than 8 bytes was made.
            DwfLibraryError: An error occurred while executing the operation.
        """
        if len(data) > 8:
            raise PyDwfError("CAN message too long.")

        # We convert the 'bytes' input data to an array of ctypes unsigned chars.
        # This silences a spurious mypy warning.

        data_ctypes_array_type = (ctypes.c_ubyte * len(data))  # pylint: disable=superfluous-parens
        data_ctypes = data_ctypes_array_type(*data)

        result = self.lib.FDwfDigitalCanTx(
                self.hdwf,
                v_id,
                extended,
                remote,
                len(data),
                data_ctypes
            )

        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def rx(self, size: int = 8) -> Tuple[int, bool, bool, bytes, int]:
        """Receive incoming CAN bus data.

        Returns:
            Tuple[int, bool, bool, bytes, int]: A tuple (vID, extended, remote, data, status)

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """

        c_v_id     = typespec_ctypes.c_int()
        c_extended = typespec_ctypes.c_int()
        c_remote   = typespec_ctypes.c_int()
        c_dlc      = typespec_ctypes.c_int()
        c_data     = ctypes.create_string_buffer(size)
        c_status   = typespec_ctypes.c_int()

        result = self.lib.FDwfDigitalCanRx(
            self.hdwf,
            c_v_id,
            c_extended,
            c_remote,
            c_dlc,
            ctypes.cast(c_data, typespec_ctypes.c_unsigned_char_ptr),
            size,
            c_status)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        v_id      = c_v_id.value
        extended = bool(c_extended.value)
        remote   = bool(c_remote.value)
        dlc      = c_dlc.value
        data     = c_data.raw[:dlc]
        status   = c_status.value

        return (v_id, extended, remote, data, status)



"""The |pydwf.core.api.device_control| module provides a single class: |Device|."""

from typing import Optional

from pydwf.core.dwf_library_subapi import AbstractDwfLibrarySubAPI

from pydwf.core.auxiliary.typespec_ctypes import typespec_ctypes
from pydwf.core.auxiliary.constants import RESULT_SUCCESS
from pydwf.core.dwf_device import DwfDevice


class DeviceControl(AbstractDwfLibrarySubAPI):
    """The |DeviceControl| class provides access to the device control functionality of a
    |DwfLibrary:link|.

    Attention:
        Users of |pydwf| should not create instances of this class directly.

        It is instantiated during initialization of a |DwfLibrary| and subsequently assigned to its
        public |deviceControl:link| attribute for access by the user.
    """

    def open(self, device_index: int, config_index: Optional[int] = None) -> DwfDevice:
        """Open a Digilent Waveforms device identified by the device index, using a specific
        device configuration index if specified.

        Note:
            This method combines the functionality of the C API functions 'FDwfDeviceOpen()' and
            'FDwfDeviceConfigOpen()' into a single method.
            The call that is actually made depends on the value of the *config_index* parameter.

        Note:
            This method can take several seconds to complete.

        Parameters:
            device_index (int): The zero-based index of the previously enumerated device (see the
                |DeviceEnum.enumerateDevices:link| method).

                To automatically enumerate all connected devices and open the first discovered device,
                use the value -1 for this parameter.

            config_index (Optional[int]): The zero-based index of the device configuration to use
                (see the |DeviceEnum.enumerateConfigurations:link| method).
                If None, open the default (first) device configuration.

        See Also:
            The |pydwf.utilities.openDwfDevice:link| convenience function provides a more powerful way to select
            and open a device and, if desired, specify its device configuration.

        Returns:
            DwfDevice: The |DwfDevice| instance created as a result of this call.

        Raises:
            DwfLibraryError: The specified device or configuration cannot be opened.
        """
        c_hdwf = typespec_ctypes.HDWF()
        if config_index is None:
            result = self.lib.FDwfDeviceOpen(device_index, c_hdwf)
        else:
            result = self.lib.FDwfDeviceConfigOpen(device_index, config_index, c_hdwf)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        hdwf = c_hdwf.value
        return DwfDevice(self.dwf, hdwf)

    def openEx(self, options: str, separator: str = ",") -> DwfDevice:
        """Open a device using options given as a string.

        This provides an extended (hence the 'Ex' prefix) version of the `:py:meth:open` method.

        The following table lists the options that can be specified in the *options* string, as listed in the DWF
        documentation. Not all of these have been verified to work.

        .. table:: Options for the openEx method
           :align: center

           +---------------------------------+-----------------------------------------------------------------+
           | option                          | description                                                     |
           +---------------------------------+-----------------------------------------------------------------+
           | index:#                         | Connect to device by enumeration 0 based index.                 |
           +---------------------------------+-----------------------------------------------------------------+
           | sn:##########                   | Open by serial number. (The devices will be enumerated).        |
           +---------------------------------+-----------------------------------------------------------------+
           | name:*device-name*              | Open by device name number. (The devices will be enumerated).   |
           +---------------------------------+-----------------------------------------------------------------+
           | config:#                        | Use configuration (0-based index).                              |
           +---------------------------------+-----------------------------------------------------------------+
           | ip:#.#.#.#/host                 | Connect to network device identified by IP address or hostname. |
           +---------------------------------+-----------------------------------------------------------------+
           | ip:*user*:*pass*@#.#.#.#/*host* | Connect to network device (with username and password).         |
           +---------------------------------+-----------------------------------------------------------------+
           | user:*username*                 | Provide username.                                               |
           +---------------------------------+-----------------------------------------------------------------+
           | pass:*password*                 | Provide password.                                               |
           +---------------------------------+-----------------------------------------------------------------+
           | secure:#                        | Enable TLS communication encryption (0/1).                      |
           +---------------------------------+-----------------------------------------------------------------+

        Todo:
            Figure out the precise syntax and semantics of options.

        Note:
            This method was added in DWF version 3.17.

        Parameters:
            options(str): A string listing options, separated by a 'separator' string (see below).
            separator(str): The separator string that separating options. Default: a single comma (',').

        Returns:
            DwfDevice: The |DwfDevice| instance created as a result of this call.

        Raises:
            DwfLibraryError: The operation could not be performed.
        """
        c_hdwf = typespec_ctypes.HDWF()

        if separator != "\n":
            options = options.replace(separator, "\n")

        options_arg = typespec_ctypes.c_const_char_ptr(options.encode())

        result = self.lib.FDwfDeviceOpenEx(options_arg, c_hdwf)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        hdwf = c_hdwf.value
        return DwfDevice(self.dwf, hdwf)

    def closeAll(self) -> None:
        """Close all Digilent Waveforms devices opened by the calling process.

        This method does not close all Digilent Waveforms devices across all processes.

        Raises:
            DwfLibraryError: The *close all* operation failed.
        """
        result = self.lib.FDwfDeviceCloseAll()
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()



"""The |pydwf| package provides classes and types to control Digilent Waveforms devices."""

# The version number of the *pydwf* package.
__version__ = "1.1.15"

# The *DwfLibrary* type is the only type users need to access library functionality.
from pydwf.core.dwf_library import DwfLibrary

# Import the 27 enumeration types and make them available for import directly from the *pydwf* package.
from pydwf.core.auxiliary.enum_types import (DwfErrorCode, DwfEnumFilter, DwfEnumConfigInfo, DwfDeviceID,
                                             DwfDeviceVersion, DwfDeviceParameter, DwfWindow, DwfState,
                                             DwfTriggerSource, DwfTriggerSlope, DwfAcquisitionMode,
                                             DwfAnalogInFilter, DwfAnalogCoupling, DwfAnalogInTriggerType,
                                             DwfAnalogInTriggerLengthCondition, DwfAnalogOutFunction,
                                             DwfAnalogOutNode, DwfAnalogOutMode, DwfAnalogOutIdle,
                                             DwfDigitalInClockSource, DwfDigitalInSampleMode,
                                             DwfDigitalOutOutput, DwfDigitalOutType, DwfDigitalOutIdle,
                                             DwfAnalogIO, DwfAnalogImpedance, DwfDmm)

# Import the two exception classes and make them available for import directly from the *pydwf* package.
from pydwf.core.auxiliary.exceptions import PyDwfError, DwfLibraryError
